#include "sprites.h"
#include "config.h"
#include "global.h"
#include <cstdlib>
#include <stdexcept>



/*
 *  Sprite
 */

Sprite::Sprite(const Pix * pix, Coord pos)
: pix_(pix)
, pos_(pos)
{
	if (!pix) throw std::runtime_error ("Null pixmap for sprite");
	Draw();
}



/*
 *  Projectile
 */

bool Projectile::Collide(Coord other, Coord size) const
{
	// Aligned rectangles intersect <=> their projections intersects
	return !(Position().x + Width() / 2  < other.x - size.x / 2 ||
	         Position().x - Width() / 2  > other.x + size.x / 2 ||
		 Position().y + Height() / 2 < other.y - size.y / 2 ||
		 Position().y - Height() / 2 > other.y + size.y / 2);
}


bool Projectile::Out() const
{
	if (Position().y < 0)
		return Position().y + Height() / 2 + 1 < 0;
	return Position().y - Height() / 2 - 1 > X11::Inst().WindowHeight();
}



/*
 *  Trajectory
 */

Trajectory::Trajectory(const RCoord * arrival,
                       bool mirrored,
                       const Coord & base_cruise,
                       Coord delta_cruise)
: stage_       (arriving)
, arrival_     (arrival)
, mirrored_    (mirrored)
, base_cruise_ (base_cruise)
, delta_cruise_(delta_cruise)
{
	if (!arrival || *arrival == RCEnd) {
		throw std::runtime_error ("Empty arrival trajectory");
	}
}


Coord Trajectory::NextPosition(Coord from, int speed)
{
	if (speed == 0) return from;

	switch (stage_) {
	case arriving: {
		const Coord goal (mirrored_ ?
		                  arrival_->XMirror() : *arrival_);
		const Coord delta (goal - from);
		double scale (delta.Length() / speed);
		if (scale < 1.0) {
			if (*(++arrival_) == RCEnd) stage_ = joining;
			return NextPosition(from, speed);
		}
		if (scale == 1.0 && *(++arrival_) == RCEnd) stage_ = joining;
		return from + delta / scale;
	}
	case joining: {
		const Coord cruise_pos (base_cruise_ + delta_cruise_);
		const Coord delta (cruise_pos - from);
		const double scale (delta.Length() / speed);
		if (scale <= 1.0) {
			stage_ = cruising;
			return cruise_pos;
		}
		else return from + delta / scale;
	}
	case cruising:
		return base_cruise_ + delta_cruise_;
	case attacking: default: {
		const Coord goal (*attack_it_);
		const Coord delta (goal - from);
		double scale (delta.Length() / (speed - 1));
		if (scale <= 1.0 && ++attack_it_ == attack_.end()) {
			stage_ = joining;
			from.y = -g_alien_height;
			return from;
		}
		if (scale < 1.0) return NextPosition(from, speed);
		return from + delta / scale;
	}
	} // switch
}


void Trajectory::BuildAttack(Coord from)
{
	const int min_delta_y (RCoord::max / 64);
	attack_.clear();
	attack_.reserve(RCoord::max / min_delta_y);
	RCoord next (from.ToRCoord());
	do {
		next += RCoord(std::rand() % (RCoord::max/3) - RCoord::max/3/2,
		               std::rand() % (RCoord::max/4) + min_delta_y);
		attack_.push_back(next);
	} while (!next.YOut());
	attack_it_ = attack_.begin();
	stage_ = attacking;
}



/*
 *  Alien
 */

Alien::Alien(const Pix * pix, const Trajectory & trajectory, int speed)
: Sprite     (pix, trajectory.InitPosition())
, trajectory_(trajectory)
, speed_     (speed)
{
}


Bullet * Alien::Fire() const
{
	Coord fire_pos (Position());
	fire_pos.y += Height() / 2;
	const int bomb_speed_x (std::rand() % 3 + std::rand() % 3 - 2);
	const Coord bomb_speed (bomb_speed_x,
	                        speed_ * 3 / 2 - std::abs(bomb_speed_x));
	return new Bullet(PixKeeper::Instance().Get(bomb_xpm),
	                  fire_pos, bomb_speed);
}
