#ifndef SPRITES_H
#define SPRITES_H

#include "gfxinterface.h"
#include "pix.h"
#include "xpms.h"
#include <vector>


// A sprite is defined by a Pix and a position.
class Sprite {
	const Pix * pix_;
	Coord pos_;
public:
	Sprite(const Pix * pix, Coord pos);
	virtual ~Sprite() {}
	Coord Position() const { return pos_; }
	int Width()  const { return pix_->Width(); }
	int Height() const { return pix_->Height(); }
	Coord Dim()  const { return pix_->Dim(); }
	void Erase() const { pix_->Erase(pos_); }
	void Draw()  const { pix_->Draw(pos_); }
protected:
	void SetPos(Coord pos) { pos_ = pos; }
	void MoveDraw(Coord delta) {
		const Coord from (pos_);
		pos_ += delta;
		pix_->Move(from, pos_);
	}
	void MoveTo(Coord to)
		{ pix_->Move(pos_, to); pos_ = to; }
};


// Projectile is the base class for constant speed moving objects.
class Projectile : public Sprite {
	const Coord speed_;
public:
	Projectile(const Pix * pix, Coord pos, Coord speed)
		: Sprite(pix, pos), speed_(speed) {}
	void Move() { MoveDraw(speed_); }
	bool Collide(Coord other, Coord size) const;
	bool Out() const;
};


// A bullet is displayed by a pix and moves straight at a constant 2D speed.
class Bullet : public Projectile {
public:
	Bullet(const Pix * pix, Coord pos, Coord speed)
		: Projectile(pix, pos, speed) {}
};


// A bonus is displayed by a pix and drops at a constant speed.
class Bonus : public Projectile {
public:
	enum bonus_t { none, extra_speed, extra_fire, extra_shield };
private:
	const bonus_t type_;
public:
	Bonus(const Pix * pix, Coord pos, bonus_t type, Coord speed = Coord(0,1))
		: Projectile(pix, pos, speed), type_(type) {}
	bonus_t Type() const { return type_; }
};


// One trajectory is given to each Alien. The trajectory computes the next
// position of the Alien. The Trajectory can also build an attack on demand.
// A trajectory has 4 stages:
// - an arriving stage consisting of passing through given RCoord's;
// - a joining stage to join the cruising position;
// - a cruising stage at position given by a changing base and a fixed delta;
// - an attacking stage consisting of passing through random RCoord's.
// The trajectory records the position of an alien.
class Trajectory {
public:
	enum stage_t { arriving, joining, cruising, attacking };
private:
	stage_t stage_;
	const RCoord * arrival_;
	const bool mirrored_;
	const Coord & base_cruise_;
	const Coord delta_cruise_;
	std::vector<RCoord> attack_;
	std::vector<RCoord>::const_iterator attack_it_;
public:
	Trajectory(const RCoord * arrival,
	           bool mirrored,
	           const Coord & base_cruise,
	           Coord delta_cruise);
	stage_t Stage() const { return stage_; }
	Coord NextPosition(Coord from, int speed);
	Coord InitPosition() const {
		return Coord(mirrored_ ? arrival_[0].XMirror() : arrival_[0]);
	}
	void BuildAttack(Coord from);
};


// An alien vessel is displayed by a pix at a position given by its trajectory.
// It can fire bullets (dropping bombs). It can collide with player bullets.
class Alien : public Sprite {
	Trajectory trajectory_;
	const int speed_;
public:
	Alien(const Pix * pix, const Trajectory & trajectory, int speed);
	void Move()
		{ MoveTo(trajectory_.NextPosition(Position(), speed_)); }
	Bullet * Fire() const;
	Trajectory::stage_t Stage() const
		{ return trajectory_.Stage(); }
	void BuildAttack()
		{ trajectory_.BuildAttack(Position()); }
};

#endif
